/* 
 * $RCSfile: irix_fddi_i.h,v $
 *
 * x-kernel v3.3
 *
 * Copyright (c) 1993 Massachusetts Board of Regents
 *
 * $Log: irix_fddi_i.h,v $
 * Revision 1.2  1996/01/30  20:48:13  slm
 * Updated version.
 *
 * Revision 1.1  1995/07/29  02:36:39  slm
 * Initial revision
 *
 * Revision 1.4.1.1.1.1  1994/11/12  19:06:38  hkaram
 * New branch
 *
 * Revision 1.4.1.1  1994/03/14  20:26:26  umass
 * Increased send and receive buffer sizes
 *
 * Revision 1.4  1993/11/13  00:45:30  menze
 * Original version from UMass
 *
 * 1.4
 * 1993/08/17 20:08:33
 */

#define IRIXFDDI_MAX_INSTANCES	   5
#define MAX_FDDI_WARM_POOL         10

#define MAX_ERROR_COUNT 1000

#define ERROR_PORT      (-1)
#define RCVBUFSIZE 	((64*1024) - 16)
#define SNDBUFSIZE 	((64*1024) - 16)
#define SMALLBUFSIZE    (14*1024)

#define HEADERPAD        RAW_HDRPAD(sizeof(struct fddi))

typedef struct
{
  struct fddi link_header;
} sendHeader;

typedef struct
{
  struct snoopheader packetInfo;
  /* char pad[HEADERPAD]; - HEADERPAD is zero for FDDI!!! */
  struct fddi link_header;
} recvHeader;

/*
 * MIN_SEND_FDDI_SIZE is larger than implied the ANSI specification, 
 * but using smaller numbers appears to bring down the ring.
 */
#define MIN_SEND_FDDI_SIZE          60
/*
 * MAX_SEND/RECV_FDDI_SIZE is from the destination MAC address through the
 * information block, inclusive.
 */
#define MAX_SEND_FDDI_SIZE         (MAX_FDDI_DATA_SIZE - \
                                    sizeof(llc802_2_Type) + \
                                    sizeof(sendHeader))
#define MAX_RECV_FDDI_SIZE         (MAX_FDDI_DATA_SIZE - \
                                    sizeof(llc802_2_Type) + \
                                    sizeof(recvHeader))
/*
 * The MTU is tuned to minimize number of socket reads in ping-pong tests
 * at packet sizes of 8000, 16000, etc.  Note that
 *
 *   450 bytes (for sim driver) = 422 bytes (for real driver) +
 *                                  8 bytes (for IRIX UDP header) +
 *                                 20 bytes (for IRIX IP header)
 *
 * This MTU also makes packets encapsulated in IRIX UDP/IP via simfddi
 * within the FDDI MTU (see RFC 1390).
 * This MTU is set for Snoop; change for Drain LATER!!!
 */
#define FDDI_MTU_SIZE   (MAX_FDDI_DATA_SIZE-422) /* put packet in 4K page */

/*
 * Frame control for asynchronous, long (48 bit) addresses, and 802.2 LLC
 */
#define FDDI_FRAME_CONTROL ((MAC_FC_ALEN|MAC_FC_LLC_FF)&~(MAC_FC_CLASS))
#define FDDI_MAC_BITS      (0)

typedef struct _PState {
    int         mtu;
    int	        sendSocket;
    int         sendPort;
    int	        recvSocket;
    int		recvPort;
    FDDIhost    myHost;
    int		errorCount;
    u_short     errorMask;
    bool        (*isValidPacket)( struct _PState *, recvHeader * );
} PState;

bool defaultIsValidPacket( PState *, recvHeader * );
bool isValidPacketOneFilter( PState *, recvHeader * );
bool isValidPacketTwoFilters( PState *, recvHeader * );
