/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: alpha_forward.s,v $
 *
 * HISTORY
 * $Log: alpha_forward.s,v $
 * Revision 1.17  1994/11/23  04:39:22  davidm
 * Added rollforward profiling.
 *
 * Revision 1.16  1994/11/22  02:47:57  davidm
 * #ifdef SCOUT_ROLLFORWARD must go after include of "scout_config.h".
 *
 * Revision 1.15  1994/11/22  02:15:33  davidm
 * Added "#ifdef SCOUT_ROLLFORWARD" bracket to make it compile when it's
 * undefined.
 *
 * Revision 1.14  1994/08/23  21:28:15  davidm
 * Added hand-crafted support for performance monitoring.
 *
 * Revision 1.13  1994/08/06  04:36:37  davidm
 * Fixed serious bug: SCOUT_SYNC_MAX_LENGTH was divided by 4.  Thus,
 * with bad enough luck, you could write an atomic sequence that
 * would not be rolled forward, yet the debugging code in
 * ScoutAtomicSequence would not catch it.  Shouldn't happeny anymore
 * now.
 *
 * Revision 1.12  1994/08/06  04:27:16  davidm
 * Instead of replaying interrupts in software, this implementation
 * lets the hardware take care of replay.  Simplifies code considerably.
 *
 * Revision 1.11  1994/07/04  18:19:39  davidm
 * (interrupt_maybe): profiling info is now kept in a structure defined
 * in libc/profil.c.
 *
 * Revision 1.10  1994/07/01  23:38:15  davidm
 * (interrupt_maybe): fixed bug in incrementing profile bin; also profiling
 * 	is now off with a profile_scale value of either 0 or 1 (for BSD
 * 	compatibility).
 *
 * Revision 1.9  1994/06/30  00:54:17  davidm
 * Added support for PC sampling based profiling.
 *
 * Revision 1.8  1994/06/11  00:02:36  menze
 * Added copyright notice
 *
 * Revision 1.7  1994/04/30  23:32:35  davidm
 * ipl_state array was on entry too small---overwriting "random" variables!
 *
 * Revision 1.6  1994/04/30  00:49:36  davidm
 * Clean up and SCOUT_TICKS_PER_CLOCK replaces hardcode constant 16.
 *
 * Revision 1.5  1994/04/24  03:41:08  davidm
 * relinquish didn't restore t10 and t11 before invoking save_and_replay.
 *
 * Revision 1.4  1994/04/23  04:03:49  davidm
 * Rollforward is now performed at max ipl level.  Otherwise, the
 * interrupted code just gets interrupted immediately again thus
 * entering an endless loop.
 *
 * Revision 1.3  1994/04/23  03:18:18  davidm
 * Added support for SCOUT_PLAY_LEDS.
 *
 * Revision 1.2  1994/04/21  07:12:02  davidm
 * (interrupt_maybe): special cased clock interrupt.  Clock interrupt
 * also resets system LEDs if SCOUT_PLAY_LEDS is defined.
 *
 */
/*
 * Code implementing rollforward.
 */
#include <regdef.h>
#include "alpha_asm.h"
#include "alpha_pal.h"
#include "scout_config.h"
#include "scout_sync_arch.h"
#include "flamingo.h"

#ifdef SCOUT_ROLLFORWARD

	.set noat
 #	.set noreorder

	.extern alpha_interrupt_handler
	.extern interrupt
	.extern scout_age

	.globl interrupt_maybe

#define STATE_OLD_PS		0x00
#define STATE_A0		0x08
#define STATE_A1		0x10
#define STATE_A2		0x18
#define STATE_PS		0x20
#define STATE_SIZE		0x28

		.data
		.align 3
real_ps:	.quad	0
		.align 5
saved_state:	.space	STATE_SIZE

/*#define PROFILE_ROLLFORWARD*/
#ifdef PROFILE_ROLLFORWARD

#define CHK_CNT		(0*8)
#define RF_CNT		(1*8)
#define INTR_RF_CNT	(2*8)

		.globl rf_profile

		.align 3
rf_profile:	.space 3*8

#endif /* PROFILE_ROLLFORWARD */

#ifdef SCOUT_PROFILE

#define PROF_BUF_OFF		0x00
#define PROF_BUF_SIZE_OFF	0x08
#define PROF_OFFSET_OFF		0x10
#define PROF_SCALE_OFF		0x18

		.extern profparams

#endif /* SCOUT_PROFILE */

	.text

.ent interrupt_maybe 0	# nesting level 0
	.frame sp IF_SIZE ra	# sp is frame pointer, IF_SIZE sized stack
				# frame, ra is return address (that's a lie)
	.mask 1 | (1<<26), 0	# reserve space for all 31 regs @ offset 0
	.fmask 0, 0		# no FP regs saved
	#
	# This is the entry point.
	#
	# When getting here, the OSF/1 PALcode has saved off the
	# interrupt frame as described in the Alpha arch. manual
	# already and IPL is set to the interrupt's level (grr...).
	#
interrupt_maybe:
	lda	sp, -IF_SIZE(sp)

	.prologue 1		# end of prologue, yes we use gp

	stq	 v0, IF_V0(sp)	# save regs clobbered by PAL_swpipl
	stq	 t0, IF_T0(sp)	# plus some work space

	/*
	 * Deal with hard clock interrupts quickly:
	 */
	cmpeq	a0, 1, v0	# is it a clock interrupt?
	blbc	v0, not_a_clock_interrupt	# no ->

#if defined(SCOUT_PROFILE) && !defined(SCOUT_PERFMON)
	.set	noreorder
	stq	t1, IF_T1(sp)	# get more scratch registers
	stq	t2, IF_T2(sp)
	stq	t3, IF_T3(sp)

	lda	v0, profparams
	ldq	t0, PROF_SCALE_OFF(v0)
	cmplt	t0, 2, t1
	blbs	t1, prof_done	# profiling turned off ->

	ldq	t1, PF_PC+IF_SIZE(sp)	# get PC sample
	ldq	t2, PROF_OFFSET_OFF(v0)
	subq	t1, t2, t3
	mulq	t0, t3, t1
	srl	t1, 16, t2
	bic	t2, 1, t2	# make 16bit aligned
	ldq	t3, PROF_BUF_SIZE_OFF(v0)
	cmplt	t2, t3, t3
	blbc	t3, prof_done	# PC sample out of sampled range ->
	ldq	t1, PROF_BUF_OFF(v0)
	addq	t1, t2, t1
	ldq_u	t2, (t1)	# increment bin
	extwl	t2, t1, t3
	mskwl	t2, t1, t2
	addl	t3, 1, t3
	inswl	t3, t1, t3
	or	t2, t3, t2
	stq_u	t2, (t1)	

prof_done:
	ldq	t1, IF_T1(sp)
	ldq	t2, IF_T2(sp)
	ldq	t3, IF_T3(sp)
	.set	reorder
	/*
	 * fall through
	 */
#endif /* defined(SCOUT_PROFILE) && !defined(SCOUT_PERFMON) */

	lda	v0, scout_age	# timer interrupt: increment scout_age
	ldq	t0, (v0)
	addq	t0, 1, t0
	stq	t0, (v0)
	and	t0, SCOUT_TICKS_PER_CLOCK-1, v0	# (scout_age % 16) == 0?
	beq	v0, play_leds	# yes -> need to do some more work
	ldq	v0, IF_V0(sp)
	ldq	t0, IF_T0(sp)
	lda	sp, IF_SIZE(sp)
	call_pal PAL_rti	# that's it---done with timer tick

play_leds:
#ifdef SCOUT_PLAY_LEDS
# ifdef SCOUT_VIRT_LEDS
	/*
	 * Reset bottom right 8 pixels on a PMAG-BA graphics board in
	 * TURBOchannel slot #3 (aka 0, on the DEC 3000/600):
	 */
	and	t0, 0x1f, v0
	bne	v0, not_a_clock_interrupt
	lda	v0, 0xfffffc01603400f8
	stq	zero, (v0)
# else
	/* reset system diagnostic LEDs on the back of the system: */
	lda	v0, 0xfffffc0000000000 | SSR
	ldl	t0, (v0)
	andnot	t0, SSR_DIAG_LED_MASK, t0
	stl	t0, (v0)
# endif /* SCOUT_VIRT_LEDS */
#endif

not_a_clock_interrupt:

#ifdef SCOUT_PERFMON

#define EV4_NTICKS	16	/* must be a power of two! */

	.data
	.globl ev4_current_time
	.globl ev4_last_time
	.globl ev4_selected_counter
ev4_current_time:	.quad	0
ev4_last_time:		.long	0
			.long	0
			.long	0
ev4_selected_counter:	.long	0

#define EV4_CURRENT_TIME	0
#define EV4_LAST_TIME		8
#define EV4_INT_COUNT		12
#define EV4_SELECTED_COUNTER	20

	.text
	.set noreorder

	cmpeq	a0, 4, v0	# is it a perfmon interrupt?
	stq	t1, IF_T1(sp)	# save t1
	blbc	v0, t1_saved	# no ->

# ifdef SCOUT_PROFILE
	/*
	 * Do performance counter based PC sampling.
	 */
	lda	v0, ev4_current_time
	stq	t2, IF_T2(sp)
	ldl	t1, EV4_SELECTED_COUNTER(v0)

	lda	v0, profparams
	stq	t3, IF_T3(sp)
	ldq	t3, PF_PC+IF_SIZE(sp)		# get PC sample
	cmpeq	a2, t1, t1			# a2 == selected_counter?
	ldq	t2, PROF_OFFSET_OFF(v0)
	ldq	t0, PROF_SCALE_OFF(v0)
	subq	t3, t2, t3			# t3 <- pc - offset
	mulq	t0, t3, t3			# get multiplication going

	cmplt	t0, 2, t0			# scale < 2?
	andnot	t1, t0, t0		# t0 = right cntr AND NOT prof. off?
	ldq	t1, PROF_BUF_OFF(v0)
	blbc	t0, perfmon_disabled		# -> nope

	ldq	t0, PROF_BUF_SIZE_OFF(v0)
	srl	t3, 16, t2			# conver fixed-point to int
	ldq	t3, IF_T3(sp)			# restore t3
	bic	t2, 1, t2			# make 16 bit aligned
	addq	t1, t2, t1			# address of bin (a short)
	cmplt	t2, t0, t0
	blbc	t0, perfmon_outofrange		# PC sample out of range ->
	ldq_u	v0, (t1)			# increment bin
	ldq	t2, IF_T2(sp)
	extwl	v0, t1, t0
	mskwl	v0, t1, v0
	addl	t0, 1, t0
	inswl	t0, t1, t0
	or	v0, t0, v0
	stq_u	v0, (t1)	
perfmon_profiling_done:
	ldq	t1, IF_T1(sp)			# restore scratch registers
	ldq	v0, IF_V0(sp)
	ldq	t0, IF_T0(sp)
	lda	sp, IF_SIZE(sp)
	call_pal PAL_rti

perfmon_disabled:
	ldq	t3, IF_T3(sp)
perfmon_outofrange:
	ldq	t2, IF_T2(sp)
	br	perfmon_profiling_done

# else /* !SCOUT_PROFILE */
	/*
	 * Use performance counters to measure event rates.
	 */
	rpcc	t1		# t1 <- cycle counter
	lda	v0, ev4_current_time
	stq	t2, IF_T2(sp)	# get another scratch register
	ldl	t2, EV4_LAST_TIME(v0)
	stl	t1, EV4_LAST_TIME(v0)	# last_time <- now
	ldq	t0, EV4_CURRENT_TIME(v0)
	subl	t1, t2, t2	# t2 <- now - last_time
	ldl	t1, EV4_SELECTED_COUNTER(v0)
	zap	t2, 0xf0, t2	# clear top 4 bytes
	cmpeq	a2, t1, t1	# a2 == ev4_selected_counter?
	addq	t2, t0, t2	# t2 <- current_time + (now - last_time)
	stq	t2, EV4_CURRENT_TIME(v0)
	s4addq	a2, v0, v0	# a2=0 -> PC0; a2=1 -> PC1
	ldq	t2, IF_T2(sp)	# restore scratch register
	ldl	t0, EV4_INT_COUNT(v0)
	blbc	t1, wrong_counter

	addl	t0, 1, t0	# ++int_count
	and	t0, EV4_NTICKS-1, t0
	stl	t0, EV4_INT_COUNT(v0)	# update interrupt counter
	beq	t0, t1_saved	# -> do full interrupt processing

wrong_counter:
	ldq	t1, IF_T1(sp)	# done with processing perfmon interrupt
	ldq	v0, IF_V0(sp)
	ldq	t0, IF_T0(sp)
	lda	sp, IF_SIZE(sp)
	call_pal PAL_rti

# endif /* !SCOUT_PROFILE */

	.set	reorder
#endif /* SCOUT_PERFMON */

	stq	 t1, IF_T1(sp)
t1_saved:
	stq	 t2, IF_T2(sp)
	stq	 t3, IF_T3(sp)
	stq	 a0, IF_T4(sp)	# save a0 in T4's slot
	stq	 t8, IF_T8(sp)
	stq	 t9, IF_T9(sp)
	stq	t10, IF_T10(sp)
	stq	t11, IF_T11(sp)
	stq	t12, IF_T12(sp)

	mov	SCOUT_MAX_IPL, a0	# disable all interrupts
	call_pal PAL_swpipl	# (old ipl in v0, clobbers a0, t0, t8-t11)
end_intr_AS:

	ldq	a0, (PF_PC+IF_SIZE)(sp)	# a0 = saved_pc
	lda	t8, interrupt_maybe	# t8 = interrupt_maybe
	lda	t9, end_intr_AS
	cmpule	t8, a0, t10		# t10 = t8 <= saved_pc
	cmpult	a0, t9, t11		# t11 = saved_pc < end_intr_AS
	and	t10, t11, t10		# t10 = saved_pc in intr_AS
	blbs	t10, perform_rollforward

	lda	t0, real_ps		# make sure v0 holds PS of
	ldq	t1, (t0)		# max-priority interrupt
	stq	zero, (t0)		# real_ps <- 0
	cmovne	t1, t1, v0		# v0 holds real PS now

	ldq	t0, IF_T0(sp)		# reload original t0 (dest_reg)
	and	t0, 3, t8
	bne	t8, norollforward	# t0 is unaligned ->

	subq	t0, 4, t0		# t0 = endPC (points to landmark)
	subq	t0, SCOUT_SYNC_MAX_LENGTH, t9	# t9 = endPC - maxlen
	cmpult	a0, t0, t8		# t8 = (saved_pc < endPC)
	cmpule	t9, a0, t9		# t9 = (endPC - MAX_LEN <= saved pc)
	and	t8, t9, t8
	blbs	t8, check_more		# endPC in range ->

chk_ill:
norollforward:
	mov	v0, a0			# a0 = my_ipl
	call_pal PAL_swpipl		# (clobbers v0, a0, t0, t8-t11)
	ldq	a0, IF_T4(sp)		# restore a0 (saved in T4's slot)
	stq	t4, IF_T4(sp)
	stq	t5, IF_T5(sp)
	stq	t6, IF_T6(sp)
	stq	t7, IF_T7(sp)

	stq	s0, IF_S0(sp)		# save remaining registers
	stq	s1, IF_S1(sp)
	stq	s2, IF_S2(sp)
	stq	s3, IF_S3(sp)
	stq	s4, IF_S4(sp)
	stq	s5, IF_S5(sp)
	stq	s6, IF_S6(sp)

	stq	a3, IF_A3(sp)
	stq	a4, IF_A4(sp)
	stq	a5, IF_A5(sp)
	stq	ra, IF_RA(sp)
	stq	$at, IF_AT(sp)

	lda	pv, alpha_interrupt_handler
	mov	sp, a3
	jsr	ra, (pv), alpha_interrupt_handler

	ldq	 v0, IF_V0(sp)
	ldq	 t0, IF_T0(sp)
	ldq	 t1, IF_T1(sp)
	ldq	 t2, IF_T2(sp)
	ldq	 t3, IF_T3(sp)
	ldq	 t4, IF_T4(sp)
	ldq	 t5, IF_T5(sp)
	ldq	 t6, IF_T6(sp)
	ldq	 t7, IF_T7(sp)
	ldq	 s0, IF_S0(sp)
	ldq	 s1, IF_S1(sp)
	ldq	 s2, IF_S2(sp)
	ldq	 s3, IF_S3(sp)
	ldq	 s4, IF_S4(sp)
	ldq	 s5, IF_S5(sp)
	ldq	 s6, IF_S6(sp)

	ldq	 a3, IF_A3(sp)
	ldq	 a4, IF_A4(sp)
	ldq	 a5, IF_A5(sp)
	ldq	 t8, IF_T8(sp)
	ldq	 t9, IF_T9(sp)
	ldq	t10, IF_T10(sp)
	ldq	t11, IF_T11(sp)
	ldq	t12, IF_T12(sp)
	ldq	 ra, IF_RA(sp)
	ldq	$at, IF_AT(sp)

	lda	sp, IF_SIZE(sp)
	call_pal PAL_rti

check_more:

#ifdef PROFILE_ROLLFORWARD
	lda	t11, rf_profile
	ldq	t10, CHK_CNT(t11)
	addq	t10, 1, t10
	stq	t10, CHK_CNT(t11)
#endif

	lda	t11, SCOUT_SYNC_LANDMARK	# t11 = LANDMARK
	ldl	t10, 0(t0)		# t10 = *endPC
	cmpeq	t10, t11, t10		# t10 = (*endPC == LANDMARK)
	blbc	t8, norollforward	# *endPC != LANDMARK ->

	# check safety of code:

	ldl	t8, (a0)		# insn = *saved_pc
	lda	t3, chk_table
	br	zero, chk_entry		# -> start checking saftey of code

chk_ok:
chk_loop:
	cmpult	a0, t0, t9		# t9 = (pc < endPC)
	blbc	t9, perform_hybrid_rollforward	# code is safe -> rollforward
	.set	noreorder	# the following is hand-crafted
chk_entry:
	zapnot	t8, 0xf, t8		# clear top 32 bits
	addq	a0, 4, a0		# ++pc
	srl	t8, 26, t9		# opCode = next >> 26
	srl	t8, 21, t10		# regA = insn >> 21
	s4addq	t9, t3, t9		# off = 4*opCode + base
	srl	t8, 16, t11		# regB = insn >> 16
	ldl	t9, (t9)		# load jump offset
	sll	t8, 43, t1		# disp = (int)next<<43 (32+11)
	and	t8, 0x1f, t12		# regC = insn & 0x1f
	srl	t8, 12, t2		# shift lit bit (insn<12>) into bit 0
	and	t11, 0x1f, t11		# regB &= 0x1f
	addq	t9, gp, t9
	ldl	t8, 0(a0)		# insn = *pc
	sra	t1, 43, t1		# disp >>= 43 (32+11)
	and	t10, 0x1f, t10		# regA &= 0x1f

	jmp	zero, (t9), chk_iop

chk_iop:
	cmpeq	t10, SCOUT_SYNC_TARGET_REG, t10	# t10 = (regA == target reg)
	cmpeq	t12, SCOUT_SYNC_TARGET_REG, t12	# t12 = (regC == target reg)
	cmpeq	t11, SCOUT_SYNC_TARGET_REG, t11	# t11 = (regB == target reg)
	or	t10, t12, t10	# t10 = (regA == 1) || (regC == 1)
	ornot	t2, t11, t11	# t11 = insn<12> || (regB != 1)
	andnot	t11, t10, t10	# t10 = (insn<12> || (B != 1)) && !(A=1||C=1)
	blbs	t10, chk_loop
	br	zero, norollforward

chk_mem:
	cmpeq	t10, SCOUT_SYNC_TARGET_REG, t10	# t10 = (regA == target reg)
	cmpeq	t11, SCOUT_SYNC_TARGET_REG, t11 # t11 = (regB == target reg)
	or	t10, t11, t10
	blbc	t10, chk_loop
	br	zero, norollforward

chk_br:	cmpeq	t10, SCOUT_SYNC_TARGET_REG, t10	# t10 = (regA == target reg)
	blbs	t10, norollforward
chk_fbr:
	blt	t1, norollforward	# disp < 0 ->
	s4addq	t1, a0, t1		# t1 = pc + disp
	cmpule	t1, t0, t1		# t1 = pc + disp <= endPC
	blbs	t1, chk_loop		# legal branch ->
	br	zero, norollforward

	.set	reorder

perform_rollforward:

#ifdef PROFILE_ROLLFORWARD
	lda t0, rf_profile
	ldq t8, INTR_RF_CNT(t0)
	addq t8, 1, t8
	stq t8, INTR_RF_CNT(t0)
#endif
	lda	t0, real_ps		# copy pre-empted interrupt's
	ldq	t8, (PF_PS+IF_SIZE)(sp)	# PS into real_ps (which is
	stq	t8, (t0)		# guaranteed not to be 0)
	br	zero, rollforward

perform_hybrid_rollforward:
	lda	t0, saved_state
	ldq	t8, (PF_PS+IF_SIZE)(sp)
	stq	t8, STATE_OLD_PS(t0)
	ldq	t9, IF_T4(sp)		# T4's slot backs up a0
	stq	t9, STATE_A0(t0)
	stq	a1, STATE_A1(t0)
	stq	a2, STATE_A2(t0)
	stq	v0, STATE_PS(t0)

	lda	a0, back_from_hybrid_rollforward
	stq	a0, IF_T0(sp)	# target reg = back_from_hybrid_rollforwar

rollforward:

#ifdef PROFILE_ROLLFORWARD
	lda t0, rf_profile
	ldq t8, RF_CNT(t0)
	addq t8, 1, t8
	stq t8, RF_CNT(t0)
#endif

	mov	SCOUT_MAX_IPL, t8
	stq	t8, (PF_PS+IF_SIZE)(sp)	# perform rollforward at max ipl

	ldq	 v0, IF_V0(sp)		# restore regs
	ldq	 t0, IF_T0(sp)
	ldq	 t1, IF_T1(sp)
	ldq	 t2, IF_T2(sp)
	ldq	 t3, IF_T3(sp)
	ldq	 t8, IF_T8(sp)
	ldq	 t9, IF_T9(sp)
	ldq	t10, IF_T10(sp)
	ldq	t11, IF_T11(sp)
	ldq	t12, IF_T12(sp)
	lda	sp, IF_SIZE(sp)		# pop stack
	call_pal PAL_rti
.end interrupt_maybe


.ent back_from_hybrid_rollforward 0	# nesting level 0
	.frame sp IF_SIZE ra	# sp is frame pointer, IF_SIZE sized stack
				# frame, ra is return address (that's a lie)
	.mask 1 | (1<<26), 0	# reserve space for all 31 regs @ offset 0
	.fmask 0, 0		# no FP regs saved
	
back_from_hybrid_rollforward:
	lda	sp, -(IF_SIZE+PF_SIZE)(sp)

	stq	 v0, IF_V0(sp)		# save some regs
	stq	 t0, IF_T0(sp)
	# a0 gets restored via PAL_rti
	stq	 t1, IF_T1(sp)
	stq	 t2, IF_T2(sp)
	stq	 t3, IF_T3(sp)
	stq	 a0, IF_T4(sp)	# save a0 in T4 slot
	stq	 t8, IF_T8(sp)
	stq	 t9, IF_T9(sp)
	stq	t10, IF_T10(sp)
	stq	t11, IF_T11(sp)
	stq	t12, IF_T12(sp)

	br	gp, 1f			# load PC into gp
1:	ldgp	gp,(gp)			# now load gp relative to insn's addr

	# re-build interrupt frame:
	stq	t0, (PF_PC+IF_SIZE)(sp)	# t0 holds return address (PC)
	stq	gp, (PF_GP+IF_SIZE)(sp)
	stq	a0, (PF_A0+IF_SIZE)(sp)
	stq	a1, (PF_A1+IF_SIZE)(sp)
	stq	a2, (PF_A2+IF_SIZE)(sp)
	lda	t0, saved_state		# restore original PS
	ldq	v0, STATE_OLD_PS(t0)
	stq	v0, (PF_PS+IF_SIZE)(sp)

	ldq	a0, STATE_A0(t0)	# restore int-handler's state
	ldq	a1, STATE_A1(t0)
	ldq	a2, STATE_A2(t0)
	ldq	v0, STATE_PS(t0)	# load PS into v0 for norollforward
	stq	a0, IF_T4(sp)		# a0 get's stamped on by PAL_swpipl
	br	zero, norollforward

.end back_from_hybrid_rollforward

	#
	# The following table maps an Alpha opcode into the
	# address of the code that checks whether the instruction
	# is safe to be rolled forward.
	#
	.rdata
chk_table:
	.gprel32 chk_ill	# 0x00 PAL (pal)
	.gprel32 chk_ill	# 0x01 Rsvd
	.gprel32 chk_ill	# 0x02 Rsvd
	.gprel32 chk_ill	# 0x03 Rsvd
	.gprel32 chk_ill	# 0x04 Rsvd
	.gprel32 chk_ill	# 0x05 Rsvd
	.gprel32 chk_ill	# 0x06 Rsvd
	.gprel32 chk_ill	# 0x07 Rsvd
	.gprel32 chk_mem	# 0x08 LDA (mem)
	.gprel32 chk_mem	# 0x09 LDAH (mem)
	.gprel32 chk_ill	# 0x0a Rsvd
	.gprel32 chk_mem	# 0x0b LDQ_U (mem)
	.gprel32 chk_ill	# 0x0c Rsvd
	.gprel32 chk_ill	# 0x0d Rsvd
	.gprel32 chk_ill	# 0x0e Rsvd
	.gprel32 chk_mem	# 0x0f STQ_U (mem)
	.gprel32 chk_iop	# 0x10 INTA (op)
	.gprel32 chk_iop	# 0x11 INTL (op)
	.gprel32 chk_iop	# 0x12 INTS (op)
	.gprel32 chk_iop	# 0x13 INTM (op)
	.gprel32 chk_ill	# 0x14 Rsvd
	.gprel32 chk_ok		# 0x15 FLTV (op)
	.gprel32 chk_ok		# 0x16 FLTI (op)
	.gprel32 chk_ok		# 0x17 FLTL (op)
	.gprel32 chk_mem	# 0x18 MISC (mem)
	.gprel32 chk_ill	# 0x19 PAL (pal)
	.gprel32 chk_ill	# 0x1a JSR (mem)
	.gprel32 chk_ill	# 0x1b PAL (pal)
	.gprel32 chk_ill	# 0x1c Rsvd
	.gprel32 chk_ill	# 0x1d PAL (pal)
	.gprel32 chk_ill	# 0x1e PAL (pal)
	.gprel32 chk_ill	# 0x1f PAL (pal)
	.gprel32 chk_ok		# 0x20 LDF (mem)
	.gprel32 chk_ok		# 0x21 LDG (mem)
	.gprel32 chk_ok		# 0x22 LDS (mem)
	.gprel32 chk_ok		# 0x23 LDT (mem)
	.gprel32 chk_ok		# 0x24 STF (mem)
	.gprel32 chk_ok		# 0x25 STG (mem)
	.gprel32 chk_ok		# 0x26 STS (mem)
	.gprel32 chk_ok		# 0x27 STT (mem)
	.gprel32 chk_mem	# 0x28 LDL (mem)
	.gprel32 chk_mem	# 0x29 LDQ (mem)
	.gprel32 chk_mem	# 0x2a LDL_L (mem)
	.gprel32 chk_mem	# 0x2b LDQ_L (mem)
	.gprel32 chk_mem	# 0x2c STL (mem)
	.gprel32 chk_mem	# 0x2d STQ (mem)
	.gprel32 chk_mem	# 0x2e STL_C (mem)
	.gprel32 chk_mem	# 0x2f STQ_C (mem)
	.gprel32 chk_br		# 0x30 BR (br)
	.gprel32 chk_fbr	# 0x31 FBEQ (br)
	.gprel32 chk_fbr	# 0x32 FBLT (br)
	.gprel32 chk_fbr	# 0x33 FBLE (br)
	.gprel32 chk_br		# 0x34 BSR (br)
	.gprel32 chk_fbr	# 0x35 FBNE (br)
	.gprel32 chk_fbr	# 0x36 FBGE (br)
	.gprel32 chk_fbr	# 0x37 FBGT (br)
	.gprel32 chk_br		# 0x38 BLBC (br)
	.gprel32 chk_br		# 0x39 BEQ (br)
	.gprel32 chk_br		# 0x3a BLT (br)
	.gprel32 chk_br		# 0x3b BLE (br)
	.gprel32 chk_br		# 0x3c BLBS (br)
	.gprel32 chk_br		# 0x3d BNE (br)
	.gprel32 chk_br		# 0x3e BGE (br)
	.gprel32 chk_br		# 0x3f BGT (br)

	# end of table

#endif /* SCOUT_ROLLFORWARD */
